import inspect
import logging
import time
import weakref
from functools import partial


logger = logging.getLogger(__name__)


class Tag(object):
    """``class`` that abstract tags and act as a simple container."""
    def __init__(self, name, valuable, ttl=-1, parent=None):
        """
        :param string name: tag name
        :param valuable: something that gives tag value, string or callable
        :param None,int ttl: Time to live for caching the tags, possible values
            are:
                * None (means no ttl)
                * -1 (means infinite ttl)
                * positive integer in seconds
        :param object obj: parent object the tag is attached to, it may be
            needed for the tag callable to process the tag
        """
        self.name = name
        self.ttl = ttl
        self.parent = parent if parent is None else weakref.proxy(parent)

        if inspect.isfunction(valuable):
            args_count = len(inspect.getargspec(valuable).args)
            if args_count > 1:
                raise Exception('Invalid number of argument for tag function')
            elif args_count == 1:
                self._value = partial(valuable, self.parent)
            else:
                self._value = valuable
        else:
            self._value = lambda: valuable

    @property
    def value(self):
        """Property that return tag value. Just return the provided value or
        call the function.

        """
        try:
            return self._value()
        except Exception:
            logger.exception('Calculating tag %s failed:', self.name)
            return None


def tag_inspector(mod, parent=None):
    """Inspect module to find tags.

    :param module mod: module to inspect

    Currently there are two ways to define a tag inside a module:

        * affect a string to a variable, the variable name will be the tag
          name
        * define a function that returns a value as a string or None (meaning
          the tag doesn't exist on the host), as you guessed the function name
          will define the tag name
    """
    tags = []
    for n, m in inspect.getmembers(mod):  # (name, member)
        # only keep strings or functions as tags
        if getattr(m, '__module__', None) != mod.__name__ or (
            n.startswith('_')):
            continue
        elif isinstance(m, (str, unicode)):
            # if string, it means it is constant, then ttl = -1
            ttl = -1
        elif inspect.isfunction(m):
            # if function take function ttl argument or set -1 by default
            ttl = getattr(m, 'ttl', -1)
        else:
            # whatever it is we don't care...
            continue

        logger.debug('Introspected %s with ttl %s.', n, ttl)

        # finally add the tag
        tags.append(Tag(n, m, ttl, parent))

    return tags


def get_tags(obj, tags, noresolve_tags):
    """Helper to get tags.

    :param obj: object with a dict ``tags`` attribute
    :param tags: List of tags to get (None mean all tags)
    :param noresolve_tags: List of tags to exclude from result
    """
    logger.debug('Tags request: %s, %s', unicode(tags), unicode(noresolve_tags))

    tags = set(tags) - set(noresolve_tags) if tags is not None else None

    if tags is None:
        tags = obj.tags.iterkeys()
    else:
        tags = tags & set(obj.tags)

    profile = time.time()
    result = dict((
        t,  # tag name
        dict(
            value=obj.tags[t].value,
            ttl=obj.tags[t].ttl,
        ),
    ) for t in tags)

    logger.debug('Profiling: %f seconds', time.time() - profile)
    logger.debug('Returning: %s', unicode(result))
    return result
